# top level class for setting LPs

import WoodiiBiomass, sys
from ModelFilter import Transporters
from ScrumPy.Util import Set

#sys.path.append('/extras')
from maintenance_costs import ATP_costs


ATPase = 'ADENOSINETRIPHOSPHATASE-RXN'


#
## LPs
#


class SetLPs:

    def __init__(self, m):
        self.m = m


    # used for bioenergetics module
    def Default(self, sup=False):  
        """post: return default lp"""
        
        lp = self.m.GetLP()
        return self.TransporterConstr(lp, sup)


    def BuildStandardLP(self, sup=False):
        """post: return default lp constraints with minimise objective"""

        lp = self.Default(sup)
        lp.SetObjective(self.m.sm.cnames)
        return lp
    

    # produces one g-DW biomass
    # Bounded option allows biomass export transporters to be bounded.
    # the lower constraint is arbitarily set to -100
    # default unit is mmol because values in mol are too small and cause glpk solver error for cobalamin synthesis
    # default growth rate changed to 0.1 as of 05/08 as it was found that 0.2 doesnt support growth on subs like formate+bicarbonate
    # maintenance costs set dynamically depending on the substrate used
    
    def Biomass(self, subs, nSource='AMMONIUM_n_tx', u=0.1,
                unitsInMol=False, bounded=False, setObj=True,
                sup=False,  mCostAuto=True, constraints={}, qDisplay=False):
        
        """pre:   subs is transporter(s) with s_tx or sf_tx, u is dilution rate hr^{-1}
                     unit is mmol.gdw^{-1}.hr^{-1} but in mol.gdw^{-1}.hr^{-1} if unitsInMol
                     if bounded: set bounds on biomass_tx instead of FixedFlux, if setObj: sum of flux minimised
                     if sup: supplments (e.g. vits) unblocked; qDisplay shows info about maintenance costs
            post: biomass LP"""

        unit = 1 if unitsInMol else 1000 #1 == mol, 1000 == mmol
        
        trans = Transporters(self.m)
        lp = self.Default(sup)
        lp = NullBiomass(self.m,lp) #block biomass transporters 
        lp.ClearFluxConstraint(nSource)
        subs = FormatSubs(self.m,subs)

        if bounded:
                fd = MakeBiomassFD(self.m,u,Bounded=True, mol_factor=unit)
                lp.SetFluxBounds(fd)
        else:
                fd = MakeBiomassFD(self.m, u, mol_factor=unit)
                lp.SetFixedFlux(fd)

         # subs can have any suffix and so can be biomass components as well as standard substrates
        for s in subs + trans['emtx']: 
            if "sf_tx" not in s:
                lp.SetFluxBounds({s:(0,None)}) # allow import of substrates only
            else:
                lp.ClearFluxConstraint(s) # for exchange reactions clear constraints

        #maintenance cost takes into account growth rate, unit, and substrate used
        if mCostAuto:
            cost = GetMaintenanceCost(u, subs, qDisplay) * unit
        else:
            Y_atp = input('enter GAM cost in ATP mol/gDW/hr: ')
            m_atp = input('enter NGAM cost in ATP mol/gDW/hr: ')        
            cost = (float(Y_atp)*u+float(m_atp)) * unit
        lp.SetFluxBounds({ATPase: (cost, None)})
        
        #for additional condition specific constraints
        lp.SetFluxBounds(constraints) 
            
        if setObj:
            lp.SetObjective(self.m.smx.cnames) #minimise total flux

        return lp

 
    # returns same LP as Biomass but also returns settings used to make LP
    # needed to dynamically return the biomassFD for biomass yield calculations
    # now out of date and divergent to Biomass function above
    def BiomassWithSettings(self, subs, nSource='AMMONIUM_n_tx', u=0.1,
                            unitsInMol=False, bounded=False, setObj=True, sup=False, constraints={}):
        """pre:   subs is subset of [s_tx and/or sf_tx], u is dilution rate hr^{-1}
                     nSource == 'reac_n_tx', unit is mmol.gdw^{-1}.hr^{-1} but in mol if unitsInMol
                     if bounded: set bounds on biomass_tx instead of FixedFlux
                     if setObj: sum of flux minimised, if sup: supplments (e.g. vits) unblocked
            post: return rv[%s] %s == LP and variousSettings"""

        rv = {}
        unit = 1 if unitsInMol else 1000 #1 == mol, 1000 == mmol
        
        trans = Transporters(self.m)
        rv['lp'] = self.Default(sup)
        rv['lp'] = NullBiomass(self.m, rv['lp']) #block biomass transporters #might not need this anymore after change to transport constraints
        rv['lp'].ClearFluxConstraint(nSource)
        subs = FormatSubs(self.m,subs)+trans['emtx']

        if bounded:
                fd = MakeBiomassFD(self.m,u,Bounded=True, mol_factor=unit)
                rv['lp'].SetFluxBounds(fd)
        else:
                fd = MakeBiomassFD(self.m, u, mol_factor=unit)
                rv['lp'].SetFixedFlux(fd)
        
        for s in subs:
            if "sf_tx" not in s:
                rv['lp'].SetFluxBounds({s:(0,None)}) # allow import of substrates only
            else:
                rv['lp'].ClearFluxConstraint(s) # for exchange reactions clear constraints

        #maintenance cost takes into account growth rate and unit used
        rv['lp'].SetFluxBounds({ATPase:(awoo_atp(u, Y_atp=Y_atp*unit, m_atp=m_atp*unit),None)})
        lp.SetFluxBounds(constraints) 

        if setObj:
            rv['lp'].SetObjective(self.m.smx.cnames) #minimise total flux

        rv['biomassFD'] = fd
        rv['unit'] = "mmol/gDW.h^{-1}" if not unitsInMol else "mol/gDW.h^{-1}"
        rv['substrates'] = subs
        rv['nitrigenSource'] = nSource
        rv['growthRate'] = u
        rv['supplementarySubstrates'] = 'yes' if sup else 'no'

        return rv
  

    
    # set default transporter constraints
    def TransporterConstr(self, lp, sup=False):
        """post: lp with default transporter constraints"""

        trans = Transporters(self.m)
        for f in trans['ftx']:  # fermentation mets
            lp.SetFluxBounds({f:(-100,0)})  # allow export

        for f in trans['hptx']:  # heterlogous products
            lp.SetFluxBounds({f:(-100,0)})  # allow export 

        for n in trans['ntx']:  # nitrogen source
            lp.SetFluxBounds({n:(-100,0)}) # allow export, need to be unconstrained along with s_tx's for growth

        for sf in trans['sftx']:  # mets which are both substrates and fermentation products
            lp.SetFluxBounds({sf:(-100,0)})  # allow export, need to be unconstrained along with s_tx's for growth 

        for em in trans['emtx']:  # essential micronutrients
            lp.SetFluxBounds({em:(-100,0)})  # allow export, need to be unconstrained along with s_tx's for growth
            
      # for b in trans['bmtx']:  # biomass mets
       #     lp.SetFixedFlux({b:0})  # block

        for b in trans['bmtx']:  # probably will stay like this
            lp.SetFluxBounds({b:(-100,0)})  

        for s in trans['stx']:  #  non-core substrates
            lp.SetFixedFlux({s:0}) # block

        if not sup:
            for s in trans['sup']: #  substrate supplements (e.g. vitamins)
                lp.SetFixedFlux({s:0}) # block

        return lp



#
## other non-class relatedfunctions
#

    
def FormatSubs(m, subs):
    """post: subs as [subs] if in model """

    if type(subs) is list:
            subs = subs
    elif type(subs) is str:
            subs = [subs]
    else:
        print("\nsubs is wrong data type\n")

    for sub in subs:
        if sub not in m.sm.cnames:
            print(sub, "- not in model")

    return subs


def MakeBiomassFD(m, u, Bounded=False, mol_factor=1000, qDisplay=False):

    # retrieves biomass fd in mmol by default
    fd = WoodiiBiomass.FluxDic(m, mol_factor=mol_factor) 
                                                                                             
    for i in fd:
        fd[i] *= u # set vals for biomass fd with growth rate

    # WARNING! if biomass is bounded it may no longer equal a gram!
    if Bounded:
        for tx in list(fd.keys()):
            fd[tx] = (-100,fd[tx]) #-100 is arbitrary

    return fd


def NullBiomass(m, lp):
    
    trans = Transporters(m)

    for tx in trans['bmtx']: # biomass transporters
        lp.SetFixedFlux({tx:0.0})

    return lp


# returns a substrate specific maintenance cost
# OR an average cost if there isn't a specific one
# Maintenance costs file in './extras/lists' directory
def GetMaintenanceCost(u, sub, qDisplay=True):
    """pre:  u is growth rate,
             sub is list of substrate transporters
       post: specific or average maintenance cost in mol/gDW/hr"""

    key = None
    for subs in ATP_costs:
        if Set.IsEqual(subs,sub):
            key = Set.Intersect(subs, sub) #gets keys in right order
            if qDisplay:
                print("maintenance cost for this substrate(s) was found")

    if key is not None:
        Y_atp = ATP_costs[tuple(key)]['Y_atp']
        m_atp = ATP_costs[tuple(key)]['m_atp']
        if qDisplay:
                print("GAM: ",  Y_atp, "mol/gDW/hr", "\nNGAM: ", m_atp, "mol/gDW/hr")
        
    else: # get an average cost 
        Y_atps = 0
        m_atps = 0
        
        for i in ATP_costs:
            m_atps += ATP_costs[i]['m_atp']
            Y_atps += ATP_costs[i]['Y_atp']

        m_atp = m_atps/len(list(ATP_costs.keys()))
        Y_atp = Y_atps/len(list(ATP_costs.keys()))

        if qDisplay:
            print("maintenance cost for this substrate(s) not found, so an average cost was used")
            print("GAM: ",  Y_atp, "mol/gDW/hr", "\nNGAM: ", m_atp, "mol/gDW/hr")
            
    rv = Y_atp*u+m_atp

    return rv














































#
##
### old superceded or out of use code




# maintenance cost units
# m_atp is in: mol.gDW^{-1}.hr^{-1}
# y_atp is in: mol.gDW^{-1}

##  average maintenance cost of 3 C1 fermentations
#m_atp = 3.7e-4 #NGAM
#Y_atp = 5.931e-2  #GAM

#Fructose maintenance costs
#m_atp = 6.1e-4 #NGAM
#Y_atp = 6.31e-2  #GAM



#awoo_atp =  lambda u, Y_atp = Y_atp, m_atp = m_atp : Y_atp*u+m_atp # u is growth rate in: hr^{-1}







"""
##m_atp = 0.37 #NGAM
##Y_atp = 59.31  #GAM
##awoo_atp = lambda u, Y_atp = Y_atp, m_atp = m_atp : Y_atp*u+m_atp # u is growth rate in: hr^{-1}

    # produces one flux unit biomass
    # Bounded option allows biomass export to be bounded.
    # the lower constraint is arbitarily set to -100
    def BiomassOLD(self, subs, nSource='AMMONIUM_n_tx', u=0.2, Bounded=False, sup=False):
        #pre: subs == [s_tx OR sf_tx], u is dilution rate (/hour)
          #  post: return biomass lp
        
        trans = Transporters(self.m)
        lp = self.Default(sup)
        lp = NullBiomass(self.m,lp) #block biomass transporters #might not need this anymore after change to transport constraints
        subs = FormatSubs(self.m,subs)+trans['emtx']
        lp.ClearFluxConstraint(nSource)
        
        for s in subs:
            if "sf_tx" not in s:
                lp.SetFluxBounds({s:(0,None)}) # allow import of substrates only
            else:
                lp.ClearFluxConstraint(s) # for exchange reactions clear constraints

        if Bounded:
                fd = MakeBiomassFD(self.m,u,Bounded=True)
                lp.SetFluxBounds(fd)
        else:
                fd = MakeBiomassFD(self.m, u)
                lp.SetFixedFlux(fd)

        lp.SetFluxBounds({ATPase:(awoo_atp(u),None)})  # mmol/g-dw.hour # consistant with mol_factor = 1,000 (i.e mmol)
        lp.SetObjective(self.m.sm.cnames)
        
        return lp

        """
